#define com_maroontress_libld_Collector_IMPORT
#define com_maroontress_libld_Garbages_IMPORT
#define com_maroontress_libld_Trace_IMPORT
#define com_maroontress_libld_trace_AllocatedTrace_IMPORT
#define com_maroontress_libld_trace_BranchedTrace_IMPORT
#define com_maroontress_libld_trace_DirectiveTrace_IMPORT
#define com_maroontress_libld_trace_ReleasedTrace_IMPORT
#define std_db_BTree_IMPORT
#define std_io_InputTube_IMPORT
#define std_lang_Chunk_IMPORT
#define std_lang_Error_IMPORT
#define std_lang_ImmutableChunk_IMPORT

#include <sys/types.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <com/maroontress/libld/Collector.h>
#include <com/maroontress/libld/Garbages.h>
#include <com/maroontress/libld/Trace.h>
#include <com/maroontress/libld/trace/AllocatedTrace.h>
#include <com/maroontress/libld/trace/BranchedTrace.h>
#include <com/maroontress/libld/trace/DirectiveTrace.h>
#include <com/maroontress/libld/trace/ReleasedTrace.h>
#include <std/db/BTree.h>
#include <std/io/InputTube.h>
#include <std/lang/Chunk.h>
#include <std/lang/Error.h>
#include <std/lang/ImmutableChunk.h>
   
static void
ImmutableChunk_set(struct ImmutableChunk *ic, const void *data, size_t size)
{
    ic->data = data;
    ic->size = size;
}

static void
ImmutableChunk_setChunk(struct ImmutableChunk *ic, struct Chunk c)
{
    ic->data = c.data;
    ic->size = c.size;
}

struct ThreadID {
    pid_t pid;
    intptr_t tid;
};

static const struct Error *
enableTrace(union BTree db, struct ThreadID *tid)
{
    struct ImmutableChunk key;

    ImmutableChunk_set(&key, tid, sizeof(*tid));
    return (BTree_del(db, key) < 0) ? Collector_IO_ERROR : NULL;
}

static const struct Error *
disableTrace(union BTree db, struct ThreadID *tid)
{
    struct ImmutableChunk key, val;

    ImmutableChunk_set(&key, tid, sizeof(*tid));
    ImmutableChunk_set(&val, NULL, 0);
    return (BTree_put(db, key, val) < 0) ? Collector_IO_ERROR : NULL;
}

static ssize_t
readTrace(union InputTube in, void *data, size_t size, size_t delta)
{
    data = (unsigned char *)data + delta;
    size -= delta;
    return InputTube_readFully(in, data, size);
}

static const struct Error *
setDirective(union InputTube in, union BTree tdb)
{
    struct DirectiveTrace c;
    struct ThreadID t;
    const struct Error *error;

    if (readTrace(in, &c, sizeof(c), sizeof(c.type)) < 0) {
        return Collector_UNEXPECTED_EOF;
    }
    t.pid = c.pid;
    t.tid = c.tid;
    switch (c.directiveType) {
    default:
	return NULL;
    case DirectiveTrace_ENABLED:
	error = enableTrace(tdb, &t);
	break;
    case DirectiveTrace_DISABLED:
	error = disableTrace(tdb, &t);
	break;
    }
    return error;
}

static int
isTraceEnabled(union BTree db, pid_t pid, intptr_t tid)
{
    struct ThreadID t;
    struct ImmutableChunk key;
    struct Chunk v;

    t.pid = pid;
    t.tid = tid;
    ImmutableChunk_set(&key, &t, sizeof(t));
    return BTree_get(db, key, &v);
}

struct Pointer {
    pid_t pid;
    intptr_t ptr;
};

static const struct Error *
putPointer(union InputTube in, union BTree db, union BTree tdb,
	   unsigned int seq)
{
    struct ImmutableChunk key, val;
    struct AllocatedTrace c, *v;
    struct Pointer p;
    size_t size;
    int s;

    if (readTrace(in, &c, sizeof(c), sizeof(c.type)) < 0) {
        return Collector_UNEXPECTED_EOF;
    }
    p.pid = c.pid;
    p.ptr = c.ptr;
    /* : AllocatedTrace_SIZEΥХեγǧ */
    size = AllocatedTrace_SIZE(c.count);
    if ((v = (struct AllocatedTrace *)alloca(size)) == NULL) {
        return Collector_NO_MEMORY;
    }
    if ((s = isTraceEnabled(tdb, p.pid, c.tid)) < 0) {
        return Collector_IO_ERROR;
    }
    c.type = (seq << 1) | !s;
    *v = c;
    if (readTrace(in, v, size, sizeof(c)) < 0) {
        return Collector_UNEXPECTED_EOF;
    }
    ImmutableChunk_set(&key, &p, sizeof(p));
    ImmutableChunk_set(&val, v, size);
    if ((s = BTree_putNoOverwrite(db, key, val)) < 0) {
        return Collector_NO_MEMORY;
    }
    return (s == 1) ? Collector_OVERLAP_CHUNK : NULL;
}

static const struct Error *
delPointer(union InputTube in, union BTree db)
{
    struct ReleasedTrace c;
    struct Pointer p;
    struct ImmutableChunk key;
    int s;

    if (readTrace(in, &c, sizeof(c), sizeof(c.type)) < 0) {
        return Collector_UNEXPECTED_EOF;
    }
    p.pid = c.pid;
    p.ptr = c.ptr;
    ImmutableChunk_set(&key, &p, sizeof(p));
    if ((s = BTree_del(db, key)) < 0) {
        return Collector_IO_ERROR;
    }
    return (s == 1) ? Collector_JUNK_POINTER : NULL;
}

static const struct Error *
forkPointerWithPID(union BTree db, pid_t pid, pid_t ppid, union BTree adb)
{
    struct ImmutableChunk key, val;
    struct Chunk k, v;
    struct AllocatedTrace *c;
    struct Pointer p;  
    int s;

    key.data = &ppid;
    key.size = sizeof(ppid);
    for (s = BTree_seqCursor(db, key, &k, &v);
         s == 0 && memcmp(key.data, &ppid, sizeof(ppid)) == 0;
         s = BTree_seq(db, &k, &v, BTree_NEXT)) {
	memcpy(&p, k.data, sizeof(p));
        p.pid = pid;
	ImmutableChunk_set(&key, &p, sizeof(p));

        if ((c = (struct AllocatedTrace *)alloca(v.size)) == NULL) {
            return Collector_NO_MEMORY;
        }
        bcopy(v.data, c, v.size);
        c->pid = pid;
	ImmutableChunk_set(&val, c, v.size);

        if ((s = BTree_put(adb, key, val)) < 0) {
            return Collector_NO_MEMORY;
	}
    }
    if (s < 0) {
	return Collector_IO_ERROR;
    }
    for (s = BTree_seq(adb, &k, &v, BTree_FIRST); s == 0;
         s = BTree_seq(adb, &k, &v, BTree_NEXT)) {
	ImmutableChunk_setChunk(&key, k);
	ImmutableChunk_setChunk(&val, v);
        if ((s = BTree_put(db, key, val)) < 0) {
            return Collector_NO_MEMORY;
	}
    }
    if (s < 0) {
	return Collector_IO_ERROR;
    }
    return NULL;
}

static const struct Error *
finalizeTraces(union BTree db, pid_t pid)
{
    struct ImmutableChunk key, val;
    struct Chunk k, v;
    struct AllocatedTrace *c;
    struct Pointer p;  
    int s;

    key.data = &pid;
    key.size = sizeof(pid);
    for (s = BTree_seqCursor(db, key, &k, &v);
         s == 0 && memcmp(key.data, &pid, sizeof(pid)) == 0;
         s = BTree_seq(db, &k, &v, BTree_NEXT)) {
        if ((c = (struct AllocatedTrace *)alloca(v.size)) == NULL) {
            return Collector_NO_MEMORY;
        }
        bcopy(v.data, c, v.size);
        p.pid = 0;
        p.ptr = c->type;
	ImmutableChunk_set(&key, &p, sizeof(p));
	ImmutableChunk_set(&val, c, v.size);
        if ((s = BTree_put(db, key, val)) < 0) {
            return Collector_NO_MEMORY;
	}
    }
    return NULL;
}

static const struct Error *
forkPointer(union InputTube in, union BTree db)
{
    static const int MODE = 0666;
    static const int FLAGS = (O_CREAT | O_RDWR);
    struct BranchedTrace c;
    union BTree adb;
    const struct Error *error;

    if (readTrace(in, &c, sizeof(c), sizeof(c.type)) < 0) {
        return Collector_UNEXPECTED_EOF;
    }
#if 1
    /*
      ꡼ץAλθ֤вᤷơ
      ץBȤ ץAȥץBƱIDˤʤ
      礬롣ץA꡼ΤƱVMA򡢥ץB
      ˲ȡߤΥ르ꥺǤϥ꡼ʤȤˤʤ롣
      н褹ˤϡץνλ­ץλˡ
      Υץγƥ󥯤򤹤٤ƥ١ȤФ褤
      
      ץλ­ƤʤΤǡҥץˡ
      λҤƱIDĥץ˳Ƥ󥯤򥬥١
      Ȥɬפ롣
    */
    if ((error = finalizeTraces(db, c.pid)) != NULL) {
	return error;
    }
#endif
    if ((adb = BTree_create(NULL, FLAGS, MODE)).impl == NULL) {
        return Collector_NO_MEMORY;
    }
    error = forkPointerWithPID(db, c.pid, c.ppid, adb);
    BTree_close(adb);
    return error;
}

struct Collector {
    union BTree db;
    union BTree tdb;
    const struct Error *error;
};

void
Collector_free(struct Collector *col)
{
    if (col == NULL)
	return;
    if (col->db.impl != NULL) {
	BTree_close(col->db);
    }
    if (col->tdb.impl != NULL) {
	BTree_close(col->tdb);
    }
    free(col);
}

struct Collector *
Collector_create(void)
{
    static const int MODE = 0666;
    static const int FLAGS = (O_CREAT | O_RDWR);
    struct Collector *col;

    if ((col = (struct Collector *)malloc(sizeof(*col))) == NULL)
	return NULL;
    col->db.impl = NULL;
    col->tdb.impl = NULL;
    col->error = NULL;

    if ((col->db = BTree_create(NULL, FLAGS, MODE)).impl == NULL
	|| (col->tdb = BTree_create(NULL, FLAGS, MODE)).impl == NULL) {
	Collector_free(col);
	return NULL;
    }
    return col;
}

static const struct Error *
loadGarbages(union BTree db, struct Garbages *g)
{
    struct Chunk k, v;
    struct AllocatedTrace *c;
    int s;

    for (s = BTree_seq(db, &k, &v, BTree_FIRST); s == 0;
         s = BTree_seq(db, &k, &v, BTree_NEXT)) {
        if ((c = (struct AllocatedTrace *)alloca(v.size)) == NULL) {
            return Collector_NO_MEMORY;
        }
        bcopy(v.data, c, v.size);
	if (c->type & 1)
	    continue;
        if (Garbages_put(g, c, v.size) < 0) {
	    return Collector_NO_MEMORY;
        }
	if (BTree_delCursor(db) != 0) {
	    return Collector_IO_ERROR;
	}
    }
    return (s == -1) ? Collector_IO_ERROR : NULL;
}

static struct Garbages *
createGarbages(struct Collector *col)
{
    struct Garbages *g;
    const struct Error *error;

    if ((g = Garbages_create()) == NULL) {
	col->error = Collector_NO_MEMORY;
	return NULL;
    }
    if ((error = loadGarbages(col->db, g)) != NULL) {
	Garbages_free(g);
	col->error = error;
	return NULL;
    }
    return g;
}

static int
clearDB(union BTree db)
{
    struct Chunk k, v;
    int s;

    for (s = BTree_seq(db, &k, &v, BTree_FIRST); s == 0;
         s = BTree_seq(db, &k, &v, BTree_NEXT)) {
	if (BTree_delCursor(db) != 0) {
	    return -1;
	}
    }
    return (s == -1) ? -1 : 0;
}

static int
clear(struct Collector *col)
{
    if (clearDB(col->db) < 0
	|| clearDB(col->tdb) < 0) {
	return -1;
    }
    return 0;
}

struct Garbages *
Collector_load(struct Collector *col, union InputTube in)
{
    unsigned int n, seq;
    const struct Error *error;

    if (clear(col) < 0) {
	col->error = Collector_IO_ERROR;
	return NULL;
    }
    for (seq = 0; InputTube_readFully(in, &n, sizeof(n)) == 0; ++seq) {
	switch (n) {
	case Trace_ALLOCATED:
            error = putPointer(in, col->db, col->tdb, seq);
            break;
	case Trace_RELEASED:
            error = delPointer(in, col->db);
            break;
	case Trace_BRANCHED:
            error = forkPointer(in, col->db);
            break;
	case Trace_DIRECTIVE:
            error = setDirective(in, col->tdb);
            break;
        default:
            error = Collector_UNKNOWN_TRACE_TYPE;
            break;
        }
        if (error != NULL) {
	    col->error = error;
	    return NULL;
	}
    }
    if (InputTube_hasError(in)) {
	col->error = Collector_UNEXPECTED_EOF;
	return NULL;
    }
    return createGarbages(col);
}

const struct Error *
Collector_getLastError(struct Collector *col)
{
    return col->error;
}

Error_DECLARE(Collector_IO_ERROR);
Error_DECLARE(Collector_NO_MEMORY);
Error_DECLARE(Collector_JUNK_POINTER);
Error_DECLARE(Collector_OVERLAP_CHUNK);
Error_DECLARE(Collector_UNEXPECTED_EOF);
Error_DECLARE(Collector_UNKNOWN_TRACE_TYPE);
