#define com_maroontress_libld_Trace_IMPORT
#define com_maroontress_libld_Tracer_IMPORT
#define com_maroontress_libld_trace_AllocatedTrace_IMPORT
#define com_maroontress_libld_trace_BranchedTrace_IMPORT
#define com_maroontress_libld_trace_DirectiveTrace_IMPORT
#define com_maroontress_libld_trace_ReleasedTrace_IMPORT

#include <sys/stat.h>
#include <sys/types.h>
#include <limits.h>
#include <dlfcn.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <com/maroontress/libld/Trace.h>
#include <com/maroontress/libld/Tracer.h>
#include <com/maroontress/libld/trace/AllocatedTrace.h>
#include <com/maroontress/libld/trace/BranchedTrace.h>
#include <com/maroontress/libld/trace/DirectiveTrace.h>
#include <com/maroontress/libld/trace/ReleasedTrace.h>

#ifndef RTLD_NEXT
#define RTLD_NEXT ((void *)-1L)
#endif

#ifndef SIZE_T_MAX
#define SIZE_T_MAX ((size_t)-1)
#endif

typedef intptr_t *Frame;
#define FP_PREV ((Frame)__builtin_frame_address(1))
#define FP_NEXT(x) ((Frame)((x)[0]))
#define IP_RETURN(x) ((x)[1])
#define FP_IS_ROOT(x) (FP_NEXT(x) == 0 || IP_RETURN(x) == 0)

Frame com_maroontress_libld_Tracer_getCurrentFP(void);

Frame __attribute__((__noinline__))
com_maroontress_libld_Tracer_getCurrentFP(void)
{
    return FP_PREV;
}

#define FP_CURRENT com_maroontress_libld_Tracer_getCurrentFP()

typedef struct {
    size_t size;
    int recursive;
    size_t pad[2];
} Pad;

static void *(*realMalloc)(size_t) = NULL;
static void (*realFree)(void *) = NULL;
static pid_t (*realFork)(void) = NULL;
static int fd = -1;

int
Tracer_open(const char *path)
{
    if (fd >= 0)
	return -1;
    return (fd = open(path, O_CREAT | O_TRUNC | O_WRONLY | O_APPEND,
		      S_IRUSR | S_IWUSR));
}

int
Tracer_close(void)
{
    int s;

    if ((s = fd) < 0)
	return -1;
    fd = -1;
    return close(s);
}

static intptr_t
getThreadID(void)
{
    /*
      tidͤȤơpthread_tͤpthread_equal(3)ǤӤ
      ʤޤȥ졼Ȥƥ¸Τ̤Υץ
      Ѥ뤳ȤǤʤǡåΥܥȥΥɥ쥹tid
      ȤѤ뤳Ȥˤ
    */
    Frame fp;

    fp = FP_CURRENT;
    for (fp = FP_NEXT(fp); !FP_IS_ROOT(fp); fp = FP_NEXT(fp)) {
	continue;
    }
    return (intptr_t)fp;
}

void
Tracer_setEnabled(int b)
{
    int s;
    struct DirectiveTrace c;

    if ((s = fd) < 0)
	return;
    c.type = Trace_DIRECTIVE;
    c.pid = getpid();
    c.tid = getThreadID();
    c.directiveType = (b) ? DirectiveTrace_ENABLED : DirectiveTrace_DISABLED;
    write(s, &c, sizeof(c));
    fsync(s);
}

static void *ldmalloc0(size_t);
static void *(*ldmalloc)(size_t) = ldmalloc0;
#ifdef linux
static char top_dlsym[4096], *ptr_dlsym = top_dlsym;
#endif

static void *
ldmalloc1(size_t n)
{
    void *p;
    Frame fp;
    intptr_t ip;
    int k;
    struct AllocatedTrace *c;

    if ((p = realMalloc(sizeof(Pad) + n)) == NULL)
	return NULL;
    ((Pad *)p)->size = n;
    ((Pad *)p)->recursive = 0;

    fp = FP_CURRENT;
    ip = IP_RETURN(fp);
    for (k = 1, fp = FP_NEXT(fp); !FP_IS_ROOT(fp) && IP_RETURN(fp) != ip;
	 ++k, fp = FP_NEXT(fp)) {
	continue;
    }
    if (!FP_IS_ROOT(fp)) {
	/*
	  A recursive call has been detected.  Maybe, write(2) is
	  not a system call, using malloc(3).
	*/
	((Pad *)p)->recursive = 1;
	p = (Pad *)p + 1;
	return p;
    }

    if ((c = alloca(AllocatedTrace_SIZE(k + 1))) == NULL) {
	realFree(p);
	return NULL;
    }
    p = (Pad *)p + 1;
    c->type = Trace_ALLOCATED;
    c->pid = getpid();
    c->ptr = (intptr_t)p;
    c->tid = (intptr_t)fp;
    c->len = n;
    c->count = k + 1;
    for (k = 0, fp = FP_CURRENT; !FP_IS_ROOT(fp); ++k, fp = FP_NEXT(fp))
        c->stack[k] = IP_RETURN(fp);
    c->stack[k] = 0;
    if (fd >= 0) {
	write(fd, c, AllocatedTrace_SIZE(k + 1));
	fsync(fd);
    }
    return p;
}

static void *
ldmalloc0(size_t n)
{
    Frame fp;
    intptr_t ip;
    int k;

    fp = FP_CURRENT;
    ip = IP_RETURN(fp);
    for (k = 1, fp = FP_NEXT(fp); !FP_IS_ROOT(fp) && IP_RETURN(fp) != ip;
	 ++k, fp = FP_NEXT(fp)) {
	continue;
    }
#ifdef linux
    if (!FP_IS_ROOT(fp)) {
	/*
	  A recursive call has been detected.  Maybe, dlsym(3) is
	  using malloc(3).
	*/
	char *p = ptr_dlsym;
	ptr_dlsym += (n + 0x0f) & ~0x0f;
	if (ptr_dlsym >= top_dlsym + sizeof(top_dlsym))
	    abort();
	return p;
    }
#endif
    if ((realMalloc = dlsym(RTLD_NEXT, "malloc")) == NULL)
	abort();
    ldmalloc = ldmalloc1;
    return ldmalloc(n);
}

static void ldfree0(void *);
static void (*ldfree)(void *) = ldfree0;

static void
ldfree1(void *p)
{
    struct ReleasedTrace c;

    if (p == NULL)
	return;
#ifdef linux
    if ((char *)p >= top_dlsym && (char *)p < top_dlsym + sizeof(top_dlsym))
	return;
#endif
    c.type = Trace_RELEASED;
    c.pid = getpid();
    c.ptr = (intptr_t)p;

    p = (Pad *)p - 1;
    if (!((Pad *)p)->recursive && fd >= 0) {
	write(fd, &c, sizeof(c));
	fsync(fd);
    }
    realFree(p);
}

static void
ldfree0(void *p)
{
    if ((realFree = dlsym(RTLD_NEXT, "free")) == NULL)
	abort();
    ldfree = ldfree1;
    ldfree(p);
}

static pid_t ldfork0(void);
static pid_t (*ldfork)(void) = ldfork0;

static pid_t
ldfork1(void)
{
    struct BranchedTrace c;
    pid_t pid;
    int fildes[2];

    if (pipe(fildes) < 0)
	return -1;
    if ((pid = realFork()) > 0) {
	c.type = Trace_BRANCHED;
	c.pid = pid;
	c.ppid = getpid();
	if (fd >= 0) {
	    write(fd, &c, sizeof(c));
	    fsync(fd);
	}
	write(fildes[1], &c, sizeof(c));
    }
    else if (pid == 0) {
	(void)read(fildes[0], &c, sizeof(c));
    }
    (void)close(fildes[0]);
    (void)close(fildes[1]);
    return pid;
}

static pid_t
ldfork0(void)
{
    if ((realFork = dlsym(RTLD_NEXT, "fork")) == NULL)
	abort();
    ldfork = ldfork1;
    return ldfork();
}

void *
malloc(size_t n)
{
    return ldmalloc(n);
}

void
free(void *p)
{
    ldfree(p);
}

void *
calloc(size_t n, size_t s)
{
    void *p;
    
    if (n > 0 && s > (SIZE_T_MAX / n))
	return NULL;
    n *= s;
    if ((p = ldmalloc(n)) == NULL)
	return NULL;
    memset(p, 0, n);
    return p;
}
    
void *
realloc(void *p, size_t n)
{
    void *q;

    if ((q = ldmalloc(n)) == NULL)
	return NULL;
    if (p == NULL)
	return q;
    memcpy(q, p, (((Pad *)p)[-1]).size);
    ldfree(p);
    return q;
}

#ifdef __FreeBSD__
void *
reallocf(void *p, size_t n)
{
    void *q;

    if ((q = realloc(p, n)) == NULL) {
	free(p);
    }
    return q;
}
#endif

pid_t
fork(void)
{
    return ldfork();
}
