#define com_maroontress_libld_Tracer_IMPORT
#define com_maroontress_testfunnel_funnel_Auditor_IMPORT
#define com_maroontress_testfunnel_funnel_LeakDetector_IMPORT
#define com_maroontress_testfunnel_funnel_StackFrame_IMPORT
#define com_maroontress_testfunnel_funnel_Status_IMPORT
#define com_maroontress_testfunnel_funnel_Workaround_IMPORT

#include <sys/param.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <err.h>
#include <errno.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <dlfcn.h>
#include <com/maroontress/libld/Tracer.h>
#include <com/maroontress/testfunnel/funnel/Auditor.h>
#include <com/maroontress/testfunnel/funnel/LeakDetector.h>
#include <com/maroontress/testfunnel/funnel/StackFrame.h>
#include <com/maroontress/testfunnel/funnel/Status.h>
#include <com/maroontress/testfunnel/funnel/Workaround.h>

static struct {
    const char *className;
    const char *methodName;
    void (*testInitialize)(void);
    void (*testMethod)(void);
    void (*testFinalize)(void);
    const char *traceFile;
    int timeout;
    int useTracer;
    pthread_mutex_t mutex;
    pthread_cond_t cond;
    int done;
} the;

/*
  exit(3)_exit(2)ѤȡХåǤʤʤ
  Τǡexit(3)ȤäƽλʤФʤʤ
  
  ƥȷ̤λơȤ֤ꡢե䶦̤ͭ
  ֤褤ʤʤʤ顢ƥоݤŬʥơɤǽλ
  ȡƥȤԤƤˤؤ餺褦˸Ƥޤ
  ˡ
*/
#define EXIT exit

void
Auditor_printStackTrace(void)
{
    StackFrame_printStackTrace(20);
}

void
Auditor_enableMemoryTrace(void)
{
    Tracer_setEnabled(1);
}

void
Auditor_disableMemoryTrace(void)
{
    Tracer_setEnabled(0);
}

void __attribute__((noreturn))
Auditor_fail(const char *file, int line, const char *s)
{
    fprintf(stderr, "%s:%d: %s#%s() failed: %s\n",
	    file, line, the.className, the.methodName, s);
    StackFrame_printStackTrace(20);
    EXIT(Status_NG);
}

static void *
run(void *arg UNUSED)
{
    if (the.useTracer && Tracer_open(the.traceFile) < 0) {
	warnx("Tracer_open() failed.");
	EXIT(Status_UNKNOWN);
    }
    if (the.testInitialize != NULL) {
	the.testInitialize();
    }
    the.testMethod();
    if (the.testFinalize != NULL) {
	the.testFinalize();
    }
    if (the.useTracer && Tracer_close() < 0) {
	warnx("Tracer_close() failed.");
	EXIT(Status_UNKNOWN);
    }
    pthread_mutex_lock(&the.mutex);
    the.done = 1;
    pthread_mutex_unlock(&the.mutex);
    pthread_cond_signal(&the.cond);
    return NULL;
}

static void
printLeaks(void)
{
    struct LeakDetector *ld;

    if ((ld = LeakDetector_create(the.traceFile)) == NULL) {
	warnx("LeakDetector_create() failed");
	EXIT(Status_UNKNOWN);
    }
    if (LeakDetector_printLeaks(ld) > 0) {
	EXIT(Status_MEMORY_LEAK_FOUND);
    }
    LeakDetector_free(ld);
}

static void __attribute__((noreturn))
audit(const char *className, const char *methodName, const char *traceFile)
{
    pthread_t thread;
    void *h;
    struct timespec stamp;

    if ((h = dlopen(className, RTLD_NOW)) == NULL) {
	warnx("%s: dlopen() failed: %s", className, dlerror());
	EXIT(Status_DYNAMIC_LINK_FAILED);
    }
    if ((the.testMethod = dlsym(h, methodName)) == NULL) {
	warnx("%s#%s: dlsym() failed: %s", className, methodName, dlerror());
	EXIT(Status_DYNAMIC_LINK_FAILED);
    }
    the.testInitialize = dlsym(h, "tf_initialize");
    the.testFinalize = dlsym(h, "tf_finalize");
    the.className = className;
    the.methodName = methodName;
    the.traceFile = traceFile;
    the.done = 0;
    if (pthread_mutex_init(&the.mutex, NULL) != 0) {
	warnx("pthread_mutex_init() failed");
	EXIT(Status_UNKNOWN);
    }
    if (pthread_cond_init(&the.cond, NULL) != 0) {
	warnx("pthread_cond_init() failed");
	EXIT(Status_UNKNOWN);
    }

    if (pthread_create(&thread, NULL, run, NULL) != 0) {
	warnx("pthread_create() failed");
	EXIT(Status_UNKNOWN);
    }
    clock_gettime(CLOCK_REALTIME, &stamp);
    stamp.tv_sec += the.timeout;
    pthread_mutex_lock(&the.mutex);
    while (!the.done) {
	if (pthread_cond_timedwait(&the.cond, &the.mutex,
				   &stamp) == ETIMEDOUT) {
	    EXIT(Status_TIMEOUT);
	}
    }
    pthread_mutex_unlock(&the.mutex);
    if (pthread_join(thread, NULL) != 0) {
	warnx("pthread_join() failed");
	EXIT(Status_UNKNOWN);
    }

    if (the.useTracer) {
	printLeaks();
    }
    (void)dlclose(h);
    EXIT(Status_OK);
}

void
Auditor_setTimeout(int sec)
{
    the.timeout = sec;
}

void
Auditor_useTracer(int b)
{
    the.useTracer = b;
}

int
Auditor_invoke(const char *className, const char *methodName)
{
    pid_t pid;
    int status, e;
    char traceFile[MAXPATHLEN];

    snprintf(traceFile, sizeof(traceFile), "/tmp/funnel.%d.log", getpid());
    fflush(NULL);
    if ((pid = fork()) < 0) {
	err(1, "fork");
    }
    if (pid == 0) {
	Workaround_childProcessStarted();
	audit(className, methodName, traceFile) ;
	/*NOTREACHED*/
    }
    if (waitpid(pid, &status, 0) < 0) {
	err(1, "waitpid");
    }
    if (the.useTracer && unlink(traceFile) < 0) {
	err(1, "unlink: %s", traceFile);
    }
    if (WIFEXITED(status)) {
	switch (e = WEXITSTATUS(status)) {
	case Status_OK:
	case Status_NG:
	case Status_DYNAMIC_LINK_FAILED:
	case Status_UNKNOWN:
	case Status_TIMEOUT:
	case Status_MEMORY_LEAK_FOUND:
	    return e;
	default:
	    errx(1, "unknown exit status: %d", e);
	}
	/*NOTREACHED*/
    }
    return (WIFSIGNALED(status) ? Status_SIGNAL : Status_UNKNOWN);
}
