#define com_maroontress_testfunnel_funnel_Display_IMPORT
#define com_maroontress_testfunnel_funnel_Event_IMPORT
#define com_maroontress_testfunnel_funnel_Recipe_IMPORT
#define com_maroontress_testfunnel_funnel_Socket_IMPORT
#define std_io_BufferedInputTube_IMPORT
#define std_io_FileDescriptor_IMPORT
#define std_io_InputTube_IMPORT
#define std_io_OutputTube_IMPORT
#define std_util_Pool_IMPORT

#include <sys/types.h>
#include <sys/param.h>
#include <netinet/in.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <err.h>
#include <com/maroontress/testfunnel/funnel/Display.h>
#include <com/maroontress/testfunnel/funnel/Event.h>
#include <com/maroontress/testfunnel/funnel/Recipe.h>
#include <com/maroontress/testfunnel/funnel/Socket.h>
#include <std/io/BufferedInputTube.h>
#include <std/io/FileDescriptor.h>
#include <std/io/InputTube.h>
#include <std/io/OutputTube.h>
#include <std/util/Pool.h>

#define DEFAULT_PORTNAME "30000"

static void
pourRecipe(union Pool pool, const char *const *av)
{
    union OutputTube out;

    if ((out = Pool_createOutputTube(pool)).impl == NULL)
	errx(1, "Pool_getOutputSteram() failed.");
    Recipe_print(out, av);
    OutputTube_close(out);
}

static void
drainRecipe(union Pool pool, union OutputTube out)
{
    union InputTube in;
    char buf[4096];
    ssize_t size;

    if ((in = Pool_createInputTube(pool)).impl == NULL)
	errx(1, "Pool_createInputSteram() failed.");
    /*
      μǤϥ򥳥ԡ뤳ȤˤʤΤǡPool_drainTo(out)
      Τ褦ʥ᥽åɤߤ
    */
    while ((size = InputTube_read(in, buf, sizeof(buf))) > 0) {
	if (OutputTube_writeFully(out, buf, size) < 0) {
	    errx(1, "OutputTube_writeFully() failed.");
	}
    }
    InputTube_close(in);
}

static void
handshake(union OutputTube out, const char *const *av)
{
    union Pool pool;
    long size;

    if ((pool = Pool_create()).impl == NULL)
	errx(1, "Pool_create() failed.");
    pourRecipe(pool, av);

    size = Pool_getSize(pool);
    size = htonl(size);
    if (OutputTube_writeFully(out, &size, sizeof(size)) < 0) {
	errx(1, "OutputTube_writeFully() failed.");
    }

    drainRecipe(pool, out);
    Pool_free(pool);
}

struct Display {
    int fd;
    union Event *lastEvent;
    union BufferedInputTube in;
    union OutputTube out;
    const char *const *testsuite;
};

static int
openSocket(const char *hostport)
{
    char *displayName, *colon;
    const char *portName;
    int fd;

    if (hostport == NULL)
	hostport = "127.0.0.1";
    if ((displayName = strdup(hostport)) == NULL)
	err(1, "strdup");
    if ((colon = strchr(displayName, ':')) != NULL) {
	*colon = 0;
	portName = colon + 1;
    }
    else {
	portName = DEFAULT_PORTNAME;
    }
    fd = Socket_open(displayName, portName);
    free(displayName);
    return fd;
}

struct Display *
Display_open(const char *hostport, const char *const *av)
{
    struct Display *dpy;
    int fd;

    if ((dpy = malloc(sizeof(struct Display))) == NULL)
	err(1, "malloc");
    fd = openSocket(hostport);
    Socket_setNoDelay(fd);

    if ((dpy->out = FileDescriptor_createOutputTube(
	     fd, FileDescriptor_NONE)).impl == NULL) {
	errx(1, "failed to create OuputTube.");
    }
    if ((dpy->in = BufferedInputTube_create(
	     FileDescriptor_createInputTube(
		 fd, FileDescriptor_NONE))).impl == NULL) {
	errx(1, "failed to create InputTube.");
    }
    dpy->fd = fd;
    dpy->lastEvent = NULL;
    dpy->testsuite = av;
    return dpy;
}

void
Display_close(struct Display *dpy)
{
    Event_free(dpy->lastEvent);
    InputTube_close(dpy->in.super.inputTube);
    OutputTube_close(dpy->out);
    close(dpy->fd);
    free(dpy);
}

void
Display_sendRecipe(struct Display *dpy)
{
    handshake(dpy->out, dpy->testsuite);
}

union Event *
Display_getEvent(struct Display *dpy)
{
    Event_free(dpy->lastEvent);
    dpy->lastEvent = Event_create(dpy, dpy->in.super.inputTube);
    return dpy->lastEvent;
}

void
Display_sendData(struct Display *dpy, const void *data, size_t size)
{
    if (OutputTube_writeFully(dpy->out, data, size) < 0) {
	errx(1, "OutputTube_writeFully() failed.");
    }
}
